#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Predict the Malware categories
"""

import os

import numpy as np
import pandas as pd

from keras.models import load_model
from keras.preprocessing import sequence

from .utils import (url2domain, get_app_file_path, download_file, find_ngrams,
                    MODELS_BASE_URL)

NGRAMS = 2
FEATURE_LEN = 128


def load_model_data(year, latest=False):
    model_fn = 'malware_cat_lstm_{0:d}.h5'.format(year)
    model_path = get_app_file_path('pydomains', model_fn)
    if not os.path.exists(model_path) or latest:
        print("Downloading Malware model data from the server ({0!s})..."
              .format(model_fn))
        if not download_file(MODELS_BASE_URL + model_fn, model_path):
            print("ERROR: Cannot download Malware model data file")
            return None, None
    else:
        print("Using cached Malware model data from local ({0!s})...".format(model_path))
    vocab_fn = 'malware_cat_vocab_{0:d}.csv'.format(year)
    vocab_path = get_app_file_path('pydomains', vocab_fn)
    if not os.path.exists(vocab_path) or latest:
        print("Downloading Malware vocab data from the server ({0!s})..."
              .format(vocab_fn))
        if not download_file(MODELS_BASE_URL + vocab_fn, vocab_path):
            print("ERROR: Cannot download Malware vocab data file")
            return None, None
    else:
        print("Using cached Malware vocab data from local ({0!s})...".format(vocab_path))
    print("Loading Malware model and vocab data file...")
    #  sort n-gram by freq (highest -> lowest)
    vdf = pd.read_csv(vocab_path)
    vocab = vdf.vocab.tolist()
    model = load_model(model_path)

    return (model, vocab)


def pred_malware(df, domain_names="domain_names", year=2017, latest=False):
    """Predict the probability of a domain being implicated in
       distributing malware using the Malware model.

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): Malware model year. Only 2017 is available.
            Default is `2017`.
        latest (Boolean): Whether or not to download latest 
            model available from GitHub. Default is `False`.

    Returns:
        DataFrame: Pandas DataFrame with 3 additional columns:
            - `pred_malware_year_domain`: the domain name
            - `pred_malware_year_lab`: most probable category
            - `pred_malware_year_prob`: probability of the domain being
              implicated in distributing malware.
    """

    if domain_names not in df.columns:
        print("No column `{0!s}` in the DataFrame".format(column))
        return None

    model, vocab = load_model_data(year, latest)

    if model is None or vocab is None:
        print("ERROR: Couldn't load model data.")
        return None

    col_domain = 'pred_malware_{0:04d}_domain'.format(year)
    col_lab =  'pred_malware_{0:04d}_lab'.format(year)
    col_prob =  'pred_malware_{0:04d}_prob'.format(year)
    df[col_domain] = df[domain_names].apply(lambda c: url2domain(c, exclude_subdomains=['www']))

    # build X from index of n-gram sequence
    X = np.array(df[col_domain].apply(lambda c: find_ngrams(vocab, c, NGRAMS)))
    X = sequence.pad_sequences(X, maxlen=FEATURE_LEN)

    df[col_lab] = model.predict_classes(X, verbose=2)
    df[col_prob] = model.predict_proba(X, verbose=2)

    return df


if __name__ == "__main__":
    pass



from .dmoz_cat import dmoz_cat
from .phish_cat import phish_cat
from .shalla_cat import shalla_cat
from .pred_phish import pred_phish
from .pred_shalla import pred_shalla
from .pred_toulouse import pred_toulouse
from .pred_malware import pred_malware

__all__ = ['dmoz_cat', 'phish_cat', 'shalla_cat',
           'pred_phish', 'pred_shalla', 'pred_toulouse', 'pred_malware']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Get the domains name category by Shalla database
"""

import os

import pandas as pd

from .utils import (url2domain, get_app_file_path, download_file,
                    DATA_BASE_URL)


def load_shalla_data(year, latest=False):
    data_fn = 'shalla_{0:d}.csv.bz2'.format(year)
    data_path = get_app_file_path('pydomains', data_fn)
    if not os.path.exists(data_path) or latest:
        print("Downloading Shallalist data from the server ({0!s})..."
              .format(data_fn))
        if not download_file(DATA_BASE_URL + data_fn, data_path):
            print("ERROR: Cannot download Shallalist data file")
            return None
    else:
        print("Using cached Shallalist data from local ({0!s})...".format(data_path))
    print("Loading Shallalist data file...")
    df = pd.read_csv(data_path, usecols=['domain', 'shalla_cat'])
    df.columns = ['__domain', 'shalla_{0:04d}_cat'.format(year)]
    return df


def shalla_cat(df, domain_names="domain_names", year=2017, latest=False):
    """Appends Shallalist domain categories to the DataFrame.

    The function extracts the domain name along with the subdomain 
    from the specified column and appends the category (dmoz_cat) 
    to the DataFrame. If Shallalist data is not available locally or 
    latest is set to `True`, it downloads the file. The function 
    looks for category of the domain name in the Shallalist file 
    for each domain. When no match is found, it returns an 
    empty string.

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): Shallalist data year. Only 2017 data is available.
            Default is 2017.
        latest (Boolean): Whether or not to download the latest 
            data available from GitHub. Default is False.

    Returns:
        DataFrame: Pandas DataFrame with two additional columns:
            'shalla_year_domain' and 'shalla_year_cat'
    """

    shalla_df = load_shalla_data(year, latest)
    if shalla_df is not None:
        try:
            col_domain = "shalla_{0:04d}_domain".format(year)
            df[col_domain] = df[domain_names].apply(lambda c:
                                                    url2domain(c))
            odf = pd.merge(df, shalla_df, how='left',
                           left_on=col_domain, right_on='__domain')
            del odf['__domain']
            return odf
        except KeyError as e:
            print("ERROR: Column {0!s} not found in the input file".format(e))
        except Exception as e:
            print("ERROR: {0!s}".format(e))

    return None



if __name__ == "__main__":
    pass



import sys
from contextlib import contextmanager
try:
    from StringIO import StringIO
except:
    from io import StringIO


@contextmanager
def capture(command, *args, **kwargs):
    out, sys.stdout = sys.stdout, StringIO()
    command(*args, **kwargs)
    sys.stdout.seek(0)
    yield sys.stdout.read()
    sys.stdout = out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Get the domain category from the PhishTank database
"""

import os

import pandas as pd

from .utils import (url2domain, get_app_file_path, download_file,
                    DATA_BASE_URL)


def load_phish_data(year, latest=False):
    data_fn = 'phish_{0:d}.csv.bz2'.format(year)
    data_path = get_app_file_path('pydomains', data_fn)
    if not os.path.exists(data_path) or latest:
        print("Downloading PhishTank data from the server ({0!s})..."
              .format(data_fn))
        if not download_file(DATA_BASE_URL + data_fn, data_path):
            print("ERROR: Cannot download PhishTank data file")
            return None
    else:
        print("Using cached PhishTank data from local ({0!s})...".format(data_path))
    print("Loading PhishTank data file...")
    df = pd.read_csv(data_path, usecols=['domain', 'verified'])
    df.drop_duplicates('domain', inplace=True)
    df.columns = ['phish_{0:04d}_cat'.format(year), '__domain']
    return df


def phish_cat(df, domain_names="domain_names", year=2017, latest=False):
    """Appends PhishTank data to the DataFrame
    
    The function extracts the domain name along with the subdomain 
    from the specified column and appends the category (phish_cat)
    to the DataFrame. If Phishtank file is not available locally or 
    latest is set to True, it downloads the file. The function 
    looks for category of the domain name in the Phishtank file 
    for each domain. When no match is found, it returns an 
    empty string.

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): PhishTank data year. Only 2017 data is available.
            Default is 2017.
        latest (Boolean): Whether or not to download latest 
            data available from GitHub. Default is `False`.

    Returns:
        DataFrame: Pandas DataFrame with two additional columns:
            'phish_year_domain' and 'phish_year_cat'
    """

    phish_df = load_phish_data(year, latest)
    if phish_df is not None:
        try:
            col_domain = "phish_{0:04d}_domain".format(year)
            df[col_domain] = df[domain_names].apply(lambda c:
                                                    url2domain(c))
            odf = pd.merge(df, phish_df, how='left',
                           left_on=col_domain, right_on='__domain')
            del odf['__domain']
            return odf
        except KeyError as e:
            print("ERROR: Column {0!s} not found in the input file".format(e))
        except Exception as e:
            print("ERROR: {0!s}".format(e))

    return None


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Predict the Phishtank categories
"""

import os

import numpy as np
import pandas as pd

from keras.models import load_model
from keras.preprocessing import sequence

from .utils import (url2domain, get_app_file_path, download_file, find_ngrams,
                    MODELS_BASE_URL)

NGRAMS = 2
FEATURE_LEN = 128


def load_model_data(year, latest=False):
    model_fn = 'phish_cat_lstm_{0:d}.h5'.format(year)
    model_path = get_app_file_path('pydomains', model_fn)
    if not os.path.exists(model_path) or latest:
        print("Downloading Phishtank model data from the server ({0!s})..."
              .format(model_fn))
        if not download_file(MODELS_BASE_URL + model_fn, model_path):
            print("ERROR: Cannot download Phishtank model data file")
            return None, None
    else:
        print("Using cached Phishtank model data from local ({0!s})...".format(model_path))
    vocab_fn = 'phish_cat_vocab_{0:d}.csv'.format(year)
    vocab_path = get_app_file_path('pydomains', vocab_fn)
    if not os.path.exists(vocab_path) or latest:
        print("Downloading Phishtank vocab data from the server ({0!s})..."
              .format(vocab_fn))
        if not download_file(MODELS_BASE_URL + vocab_fn, vocab_path):
            print("ERROR: Cannot download Phishtank vocab data file")
            return None, None
    else:
        print("Using the cached Phishtank vocab data from local ({0!s})...".format(vocab_path))
    print("Loading the Phishtank model and vocab data file...")
    #  sort n-gram by freq (highest -> lowest)
    vdf = pd.read_csv(vocab_path)
    vocab = vdf.vocab.tolist()
    model = load_model(model_path)

    return (model, vocab)


def pred_phish(df, domain_names="domain_names", year=2016, latest=False):
    """Predict whether domain is implicated in phishing
       using a model based on the PhishTank data.

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): PhishTank model year. Only 2016 is available.
            Default is `2016`.
        latest (Boolean): Whether or not to download latest 
            model available from GitHub. Default is `False`.

    Returns:
        DataFrame: Pandas DataFrame with additional columns:
            - `pred_phish_year_domain`: domain name
            - `pred_phish_year_lab`: most probable category
            - `pred_phish_year_prob`: probability of the domain being
              implicated in phishing.
    """

    if domain_names not in df.columns:
        print("No column `{0!s}` in the DataFrame".format(column))
        return None

    model, vocab = load_model_data(year, latest)

    if model is None or vocab is None:
        print("ERROR: Couldn't load model data.")
        return None

    col_domain = 'pred_phish_{0:04d}_domain'.format(year)
    col_lab =  'pred_phish_{0:04d}_lab'.format(year)
    col_prob =  'pred_phish_{0:04d}_prob'.format(year)
    df[col_domain] = df[domain_names].apply(lambda c: url2domain(c))

    # build X from index of n-gram sequence
    X = np.array(df[col_domain].apply(lambda c: find_ngrams(vocab, c, NGRAMS)))
    X = sequence.pad_sequences(X, maxlen=FEATURE_LEN)

    df[col_lab] = model.predict_classes(X, verbose=2)

    df[col_prob] = model.predict_proba(X, verbose=2)

    return df


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Predict the Toulouse list categories
"""

import os

import numpy as np
import pandas as pd

from keras.models import load_model
from keras.preprocessing import sequence

from .utils import (url2domain, get_app_file_path, download_file, find_ngrams,
                    MODELS_BASE_URL)

NGRAMS = 2
FEATURE_LEN = 128


def load_model_data(year, latest=False):
    model_fn = 'toulouse_cat_lstm_others_{0:d}.h5'.format(year)
    model_path = get_app_file_path('pydomains', model_fn)
    if not os.path.exists(model_path) or latest:
        print("Downloading Toulouse model data from the server ({0!s})..."
              .format(model_fn))
        if not download_file(MODELS_BASE_URL + model_fn, model_path):
            print("ERROR: Cannot download Toulouse model data file")
            return None, None, None
    else:
        print("Using cached Toulouse model data from local ({0!s})...".format(model_path))
    vocab_fn = 'toulouse_cat_vocab_others_{0:d}.csv'.format(year)
    vocab_path = get_app_file_path('pydomains', vocab_fn)
    if not os.path.exists(vocab_path) or latest:
        print("Downloading Toulouse vocab data from the server ({0!s})..."
              .format(vocab_fn))
        if not download_file(MODELS_BASE_URL + vocab_fn, vocab_path):
            print("ERROR: Cannot download Toulouse vocab data file")
            return None, None, None
    else:
        print("Using cached Toulouse vocab data from local ({0!s})...".format(vocab_path))
    names_fn = 'toulouse_cat_names_others_{0:d}.csv'.format(year)
    names_path = get_app_file_path('pydomains', names_fn)
    if not os.path.exists(names_path) or latest:
        print("Downloading Toulouse names data from the server ({0!s})..."
              .format(names_fn))
        if not download_file(MODELS_BASE_URL + names_fn, names_path):
            print("ERROR: Cannot download Toulouse names data file")
            return None, None, None
    else:
        print("Using cached Toulouse names data from local ({0!s})...".format(names_path))
    print("Loading Toulouse model, vocab and names data file...")
    #  sort n-gram by freq (highest -> lowest)
    vdf = pd.read_csv(vocab_path)
    vocab = vdf.vocab.tolist()
    model = load_model(model_path)
    cdf = pd.read_csv(names_path)
    categories = cdf.toulouse_cat.tolist()
    return (model, vocab, categories)


def pred_toulouse(df, domain_names="domain_names", year=2017, latest=False):
    """Predict the category of content hosted by the domain
     using the Toulouse model

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): Toulouse model year. Only 2017 is available.
            Default is `2017`.
        latest (Boolean): Whether or not to download latest 
            model available from GitHub. Default is `False`.
    
    Returns:
        DataFrame: Pandas DataFrame with additional columns:
            - `pred_toulouse_year_domain`: domain name
            - `pred_toulouse_year_lab`: most probable category
            - `pred_toulouse_year_prob_catname`: probability that the 
              domain hosts content of category `catname`
    """

    if domain_names not in df.columns:
        print("No column `{0!s}` in the DataFrame".format(column))
        return None

    model, vocab, cats = load_model_data(year, latest)

    if model is None or vocab is None or cats is None:
        print("ERROR: Couldn't load model data.")
        return None

    col_domain = 'pred_toulouse_{0:04d}_domain'.format(year)
    col_lab =  'pred_toulouse_{0:04d}_lab'.format(year)
    col_probs =  ['pred_toulouse_{0:04d}_prob_{1:s}'.format(year, c)
                  for c in cats]

    df[col_domain] = df[domain_names].apply(lambda c: url2domain(c, exclude_subdomains=['www']))

    # build X from index of n-gram sequence
    X = np.array(df[col_domain].apply(lambda c: find_ngrams(vocab, c, NGRAMS)))
    X = sequence.pad_sequences(X, maxlen=FEATURE_LEN)

    df['toulouse_cat'] = model.predict_classes(X, verbose=2)

    df[col_lab] = df.toulouse_cat.apply(lambda c: cats[c])
    del df['toulouse_cat']

    proba = model.predict_proba(X, verbose=2)

    pdf = pd.DataFrame(proba, columns=col_probs)
    pdf.set_index(df.index, inplace=True)

    df = pd.concat([df, pdf], axis=1)

    return df


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility functions
"""

import sys
import os
from os import path

import tldextract
from tqdm import tqdm
import requests

DATA_BASE_URL = (os.environ.get('PYDOMAINS_DATA_URL') or
                 'https://raw.githubusercontent.com/themains/pydomains/master/pydomains/data/')

MODELS_BASE_URL = (os.environ.get('PYDOMAINS_MODELS_URL') or
                  'https://raw.githubusercontent.com/themains/pydomains/master/pydomains/models/')


def isstring(s):
    # if we use Python 3
    if (sys.version_info[0] >= 3):
        return isinstance(s, str)
    # we use Python 2
    return isinstance(s, basestring)


def find_ngrams(vocab, text, n):
    """Find and return list of the index of n-grams in the vocabulary list.

    Generate the n-grams of the specific text, find them in the vocabulary list
    and return the list of index have been found.

    Args:
        vocab (:obj:`list`): Vocabulary list.
        text (str): Input text
        n (int): N-grams

    Returns:
        list: List of the index of n-grams in the vocabulary list.

    """

    wi = []

    if not isstring(text):
        return wi

    a = zip(*[text[i:] for i in range(n)])
    for i in a:
        w = ''.join(i)
        try:
            idx = vocab.index(w)
        except Exception as e:
            idx = 0
        wi.append(idx)
    return wi


def url2domain(url, exclude_subdomains=False):
    """Extract the domain from URL.
    """
    tld = tldextract.extract(url)
    a = []
    if tld.subdomain != '':
        if isinstance(exclude_subdomains, list):
            if tld.subdomain not in exclude_subdomains:
                a.append(tld.subdomain)
        elif not exclude_subdomains :
            a.append(tld.subdomain)
    a.append(tld.domain)
    if tld.suffix != '':
        a.append(tld.suffix)
    domain = '.'.join(a)
    return domain


def get_app_file_path(app_name, filename):
    user_dir = path.expanduser('~')
    app_data_dir = path.join(user_dir, '.' + app_name)
    if not path.exists(app_data_dir):
        os.makedirs(app_data_dir)
    file_path = path.join(app_data_dir, filename)
    return file_path


def download_file(url, target):

    if 'PYDOMAINS_AUTH_TOKEN' in os.environ:
        auth_token = os.environ['PYDOMAINS_AUTH_TOKEN']
        headers = {'Authorization': 'token {0!s}'.format(auth_token)}
    else:
        headers = {}

    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True, headers=headers)

    if r.status_code == 200:
        chunk_size = (64 * 1024)
        # Total size in bytes.
        total_size = int(r.headers.get('content-length', 0)) / chunk_size

        total_size += 1

        with open(target, 'wb') as f:
            for data in tqdm(r.iter_content(chunk_size), total=total_size, unit_scale=chunk_size/1024, unit='KB'):
                f.write(data)
        return True
    else:
        print("ERROR: status_code={0:d}".format(r.status_code))
        return False


if __name__ == "__main__":
    print(url2domain('http://test.hello.com/hello/test?hello=test&test=hello'))
    # PYDOMAINS_AUTH_TOKEN env. var. requires for private repo.
    # https://github.com/settings/tokens
    download_file('https://raw.githubusercontent.com/soodoku/pydomains/master/pydomains/data/dmoz_2016.csv.bz2', 'test.bz2')
    print(get_app_file_path('pydomains', 'dmoz_2016.csv.bz2'))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tests for pred_toulouse.py

"""

import os
import shutil
import unittest
import pandas as pd
from pydomains import pred_toulouse
from . import capture


class TestPredToulouse(unittest.TestCase):

    def setUp(self):
        domains = [{'label': 'test1', 'url': 'http://www.google.com'},
                 {'label': 'test2', 'url': 'http://www.sanook.com'}]
        self.df = pd.DataFrame(domains)

    def tearDown(self):
        pass

    def test_pred_toulouse_2017(self):
        odf = pred_toulouse(self.df, 'url')
        self.assertIn('pred_toulouse_2017_lab', odf.columns)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Get the category of content hosted by the domain according to the DMOZ database
"""

import os

import pandas as pd

from .utils import (url2domain, get_app_file_path, download_file,
                    DATA_BASE_URL)


def load_dmoz_data(year, latest=False):
    data_fn = 'dmoz_{0:d}.csv.bz2'.format(year)
    data_path = get_app_file_path('pydomains', data_fn)
    if not os.path.exists(data_path) or latest:
        print("Downloading DMOZ data from the server ({0!s})..."
              .format(data_fn))
        if not download_file(DATA_BASE_URL + data_fn, data_path):
            print("ERROR: Cannot download DMOZ data file")
            return None
    else:
        print("Using cached DMOZ data from local ({0!s})...".format(data_path))
    print("Loading DMOZ data file...")
    df = pd.read_csv(data_path, usecols=['domain', 'cat_labels_en', 'rdomain'])
    df.columns = ['__domain', 'dmoz_{0:04d}_cat'.format(year), '__rdomain']
    return df


def dmoz_cat(df, domain_names="domain_names", year=2016, latest=False):
    """Appends DMOZ domain categories to the DataFrame.
    
    The function extracts the domain name along with the subdomain 
    from the specified column and appends the category (dmoz_cat) 
    to the DataFrame. If DMOZ file is not available locally or 
    latest is set to True, it downloads the file. The function 
    looks for category of the domain name in the DMOZ file 
    for each domain. When no match is found, it returns an 
    empty string.

    Args:
        df (:obj:`DataFrame`): Pandas DataFrame. No default value.
        domain_names (str): Column name of the domain in DataFrame. 
            Default in `domain_names`.
        year (int): DMOZ data year. Only 2016 data is available.
            Default is 2016.
        latest (Boolean): Whether or not to download latest 
            data available from GitHub. Default is False.

    Returns:
        DataFrame: Pandas DataFrame with two additional columns:
            'dmoz_year_domain' and 'dmoz_year_cat'
    """

    dmoz_df = load_dmoz_data(year, latest)
    if dmoz_df is not None:
        try:
            col_domain = "dmoz_{0:04d}_domain".format(year)
            col_cat = "dmoz_{0:04d}_cat".format(year)
            df[col_domain] = df[domain_names].apply(lambda c:
                                                    url2domain(c))
            dmoz_cols = ['__domain', col_cat]
            odf = pd.merge(df, dmoz_df[dmoz_cols], how='left',
                           left_on=col_domain, right_on='__domain')
            del odf['__domain']
            dmoz_df.drop_duplicates('__rdomain', inplace=True)
            dmoz_cols = ['__rdomain', col_cat]
            odf = pd.merge(odf, dmoz_df[dmoz_cols], how='left',
                           left_on=col_domain, right_on='__rdomain')
            del odf['__rdomain']
            cat_xy = [col_cat + '_x', col_cat + '_y']
            odf[col_cat] = odf[cat_xy].apply(lambda r: r[0]
                                             if pd.isnull(r[1]) else r[1],
                                             axis=1)
            odf.drop(cat_xy, axis=1, inplace=True)
            return odf
        except KeyError as e:
            print("ERROR: Column {0!s} not found in the dataframe".format(e))
        except Exception as e:
            print("ERROR: {0!s}".format(e))

    return df


if __name__ == "__main__":
    pass
